#
# $HeadURL: https://svn.oucs.ox.ac.uk/sysdev/src/packages/c/configtool/tags/2.12-1~ox.0/lib/Configtool/Users.pm $
# $LastChangedRevision: 16104 $
# $LastChangedDate: 2009-08-28 17:31:39 +0100 (Fri, 28 Aug 2009) $
# $LastChangedBy: tom $
#
package Configtool::Users;

use strict;
use warnings FATAL => 'all';

use base 'Exporter';

our @EXPORT      = ();
our @EXPORT_OK   = qw( sync_userskel update );
our %EXPORT_TAGS = ();

use Config::General;
use Configtool::Util::DropPrivs;
use File::Temp;
use IO::File;
use List::MoreUtils qw( uniq );
use User::grent;
use User::pwent;
use Readonly;
#use Smart::Comments;

Readonly my $ADDUSER  => '/usr/sbin/adduser';
Readonly my $ADDGROUP => '/usr/sbin/addgroup';
Readonly my $USERMOD  => '/usr/sbin/usermod';
Readonly my $TAR      => '/bin/tar';
Readonly my $HOME     => '/home';

my $verbose;
my $dryrun;

sub _verbose {
    if ( $verbose ) {
        print "$_\n" for @_;
    }
}

sub sync_userskel {
    my ( $config ) = @_;

    my $class = $config->userskel_class;

    eval "require $class"
        or die "Failed to load $class\n";

    $class->sync_userskel( $config );
}

sub update {
    my ( $config, @wanted ) = @_;

    $verbose = $config->verbose;
    $dryrun  = $config->test;

    my $users = _get_users_config( $config->users_conf );

    if ( @wanted ) {
        my @filtered_users;
        foreach my $username ( @wanted ) {
            push @filtered_users, grep { $_->{ name } eq $username } @{ $users }
                or die "User '$username' not configured";
        }
        $users = \@filtered_users;
    }

    # Ensure all referenced groups exist
    foreach my $group ( uniq map @{ $_->{ group } || [] }, @{ $users } ) {
        if ( defined( my $g = getgrnam( $group ) ) ) {
            _verbose( "Group '$group' exists with gid " . $g->gid );
        }
        else {
            _add_group( $group );
        }
    }

    # Ensure all users exist and are in the specified supplementary groups
    foreach my $user ( @${ users } ) {
        ###  $user
        if ( defined( my $u = getpwnam( $user->{ name } ) ) ) {
            _verbose( "User '$user->{ name }' exists with uid " . $u->uid );
            _populate_user_home( $user->{ name }, $config->userskel )
                if $config->force;
        }
        else {
            _add_user( $user );
            _populate_user_home( $user->{ name }, $config->userskel );
        }
        _check_supplementary_groups( $user->{ name }, $user->{ group } );
    }

}

sub _get_users_config {
    my ( $config_file ) = @_;

    my $config = Config::General->new( $config_file )
        or die "Error reading users config: $config_file";

    my %c = $config->getall();

    my @users;

    while ( my ( $user_name, $user_properties ) = each %{ $c{ User } } ) {
        my $class_properties = {};
        if ( my $class_name = $user_properties->{ class } ) {
            $class_properties = $c{ Class }->{ $class_name }
                or die "User '$user_name' references undefined class '$class_name'";
        }
        push @users, { %{ $class_properties },
                       gecos => $user_name,
                       %{ $user_properties },
                       name => $user_name,
                   };
    }

    return \@users;
}

sub _add_group {
    my ( $name ) = @_;
    _verbose( "Adding group '$name'" );
    unless ( $dryrun ) {
        system( $ADDGROUP, '--system', $name ) == 0
            or die "addgroup $name failed";
    }
}

sub _add_user {
    my ( $user ) = @_;
    _verbose( "Adding user '$user->{ name }'" );
    return if $dryrun;
    my @opts;
    foreach my $o ( qw( home shell uid firstuid lastuid gecos ) ) {
        push @opts, "--$o", $user->{ $o }
            if defined $user->{ $o };
    }
    foreach my $o ( qw( system no-create-home disabled-password disabled-login ) ) {
        push @opts, "--$o"
            if exists $user->{ $o };
    }
    system( $ADDUSER, @opts, $user->{ name } ) == 0
        or die "adduser $user->{ name } failed";
    if ( exists $user->{ system } and exists $user->{ 'disabled-password' } ) {
        # work around adduser bug
        _verbose( "Enabling logins for system user '$user->{ name }' with disabled password" );
        system( $USERMOD, '-p', '*', $user->{ name } ) == 0
            or die "failed to modify user $user->{ name }";
    }
}

sub _add_user_group {
    my ( $user, $group ) = @_;
    _verbose( "Adding '$user' to group '$group'" );
    return if $dryrun;
    system( $USERMOD, '-a', '-G', $group, $user ) == 0
        or die "failed to add '$user' to group '$group'";
}

sub _populate_user_home {
    my ( $user, $userskel ) = @_;

    my $this_user_skel = File::Spec->catfile( $userskel, "$user.tar.gz" );
    unless ( -f $this_user_skel ) {
        _verbose( "No userskel for $user" );
        return;
    }

    _verbose( "Populating home directory for '$user' from $this_user_skel" );
    return if $dryrun;

    my $ifh = IO::File->new( $this_user_skel, O_RDONLY )
        or die "open $this_user_skel: $!";

    defined( my $pid = fork())
        or die "fork failed: $!";

    if ( $pid == 0 ) {
        drop_privs( $user );
        open( STDIN, '<&' . $ifh->fileno )
            or die "dup STDIN: $!";
        my @tar = ( $TAR, '-C', $HOME, '-z', '-x' );
        push @tar, '-v'
            if $verbose;
        exec( @tar )
            or die "exec $TAR: $!";
    }
    waitpid( $pid, 0 );
    die "failed to extract $this_user_skel (exit $?)"
        if $?;
}

sub _check_supplementary_groups {
    my ( $user, $groups ) = @_;

    unless ( defined $groups and @{ $groups } ) {
        _verbose( "No supplementary groups for '$user'" );
        return;
    }

    _verbose( "Checking supplementary groups for '$user'" );

    foreach my $group ( @{ $groups } ) {
        my $grent = getgrnam( $group )
            or die "No such group '$group'";
        if ( grep { $user eq $_ } @{ $grent->members } ) {
            _verbose( "'$user' is a member of group '$group'" );
        }
        else {
            _add_user_group( $user, $group );
        }
    }
}

1;

__END__
